package dev.fitko.fitconnect.client;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.model.SubmissionPayload;
import dev.fitko.fitconnect.client.strategies.SendEncryptedSubmissionStrategy;
import dev.fitko.fitconnect.client.strategies.SendNewSubmissionStrategy;
import dev.fitko.fitconnect.client.strategies.SubmitStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import static dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType.APPLICATION_JSON;
import static dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType.APPLICATION_XML;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildDataPayload;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildEncryptedDataPayload;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildServiceType;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.toAttachmentPayloads;

/**
 * A fluent client for announcing and handing in a {@link SubmitSubmission}
 */
public final class SenderClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(SenderClient.class);

    private SenderClient() {
    }

    public static Start build(final Sender sender, final String metadataSchema, final Environment environment) {
        return new ClientBuilder(sender, metadataSchema, environment);
    }

    public interface Start {

        /**
         * Creates new submission request.
         *
         * @return next step to add (optional) attachments
         */
        WithAttachments newSubmission();

        /**
         * Creates new submission request wit already encrypted data.
         *
         * @return next step to add (optional) attachments
         */
        WithEncryptedAttachments newSubmissionWithEncryptedData();

        /**
         * Retrieve the public encryption key
         *
         * @param destinationId unique identifier of a {@link Destination}
         * @return optional string containing the public JWK
         */
        Optional<String> getPublicKey(UUID destinationId);
    }

    public interface WithDestination {

        /**
         * Configures the client for the given destination and loads the public key
         *
         * @param destinationId unique identifier of the clients destination
         * @return the upload step for attachments
         */
        WithServiceType withDestination(UUID destinationId);
    }

    public interface WithAttachments {

        /**
         * Sends the submission with a list of attachments
         *
         * @param attachments that are sent with the submission
         * @return the next step where additional data can be added to the submission
         */
        WithData withAttachments(List<File> attachments);


        /**
         * Sends the submission with an attachments
         *
         * @param attachment that is sent with the submission
         * @return the next step where additional data can be added to the submission
         */
        WithData withAttachment(File attachment);


        /**
         * XML data as string.
         *
         * @param data json string
         * @return next step to submit the data
         */
        WithDestination withJsonData(String data);

        /**
         * XML data as string.
         *
         * @param data xml string
         * @return next step to submit the data
         */
        WithDestination withXmlData(String data);
    }

    public interface WithEncryptedAttachments {

        /**
         * Sends the submission with a list of already encrypted attachments
         *
         * @param encryptedAttachments map of encrypted attachments on attachment UUIDs
         *
         * @return the next step where additional encrypted data can be added to the submission
         */
        WithEncryptedData withEncryptedAttachments(Map<UUID, String> encryptedAttachments);

        /**
         * Sends the submission with an attachments
         *
         * @param attachmentId unique identifier attachment that is uploaded
         * @param attachment encrypted attachment that is sent with the submission
         * @return the next step where additional data can be added to the submission
         */
        WithEncryptedData withEncryptedAttachment(UUID attachmentId, String attachment);

        /**
         * JSON data as encrypted JWE string.
         *
         * @param encryptedData encrypted json string
         * @return next step to submit the data
         */
        WithEncryptedMetadata withEncryptedJsonData(String encryptedData);

        /**
         * XML data as encrypted JWE string.
         *
         * @param encryptedData encrypted xml string
         * @return next step to submit the data
         */
        WithEncryptedMetadata withEncryptedXmlData(String encryptedData);
    }

    public interface WithServiceType {

        /**
         * Sets a {@link ServiceType} for a submission
         *
         * @param name     a name of the service
         * @param leikaKey unique identifier of the service in the form <b>urn:de:fim:leika:leistung:</b>
         * @return next step to add attachments
         * @see <a href="https://fimportal.de/">Search For Service Types</a>
         */
        Submit withServiceType(String name, String leikaKey);

        /**
         * Sets a {@link ServiceType} for a submission
         *
         * @param name        a name of the service
         * @param description description of the service
         * @param leikaKey    unique identifier of the service in the form <b>urn:de:fim:leika:leistung:</b>
         * @return next step to add attachments
         * @see <a href="https://fimportal.de/">Search For Service Types</a>
         */
        Submit withServiceType(String name, String description, String leikaKey);
    }

    public interface WithEncryptedData {

        /**
         * JSON data as encrypted JWE string.
         *
         * @param encryptedData encrypted json string
         * @return next step to add encrypted metadata
         */
        WithEncryptedMetadata withEncryptedJsonData(String encryptedData);

        /**
         * XML data as encrypted JWE string.
         *
         * @param encryptedData encrypted xml string
         * @return next step to add encrypted metadata
         */
        WithEncryptedMetadata withEncryptedXmlData(String encryptedData);
    }

    public interface WithEncryptedMetadata {

        /**
         * JSON data as encrypted JWE string.
         *
         * @param encryptedMetadata encrypted JWE string of the metadata
         * @return next step to add a destination
         */
        WithDestination withEncryptedMetadata(String encryptedMetadata);
    }

    public interface WithData {

        /**
         * XML data as string.
         *
         * @param data json string
         * @return next step to submit the data
         */
        WithDestination withJsonData(String data);

        /**
         * JSON data as string.
         *
         * @param data xml string
         * @return next step to submit the data
         */
        WithDestination withXmlData(String data);

    }

    public interface Submit {

        /**
         * Send submission with attachments and data to FIT-Connect API.
         *
         * @return {@link SubmitSubmission} object of the handed in submission
         */
        Optional<SubmitSubmission> submit();
    }

    /**
     * ClientBuilder that implements all steps to guide through the API calls
     */
    public static class ClientBuilder implements
            Start,
            WithAttachments, WithEncryptedAttachments,
            WithData, WithEncryptedData,
            WithEncryptedMetadata,
            WithDestination,
            WithServiceType,
            Submit {

        private final Sender sender;
        private final Environment environment;
        private final String metadataSchema;

        private SubmitStrategy submitStrategy;
        private SubmissionPayload submissionPayload;

        public ClientBuilder(final Sender sender, final String metadataSchema, final Environment environment) {
            this.sender = sender;
            this.metadataSchema = metadataSchema;
            this.environment = environment;
            this.submissionPayload = new SubmissionPayload();
        }

        @Override
        public WithAttachments newSubmission() {
            this.submitStrategy = new SendNewSubmissionStrategy(sender, metadataSchema, environment);
            return this;
        }

        @Override
        public WithEncryptedAttachments newSubmissionWithEncryptedData() {
            this.submitStrategy = new SendEncryptedSubmissionStrategy(sender);
            return this;
        }

        @Override
        public Optional<String> getPublicKey(final UUID destinationId) {
            try {
                return retrievePublicKey(destinationId);
            } catch (final RestApiException | KeyNotRetrievedException e) {
                LOGGER.error("Public encryption key could not be retrieved", e);
            }
            return Optional.empty();
        }

        @Override
        public WithServiceType withDestination(final UUID destinationId) {
            if (destinationId == null) {
                LOGGER.error("DestinationId was null");
                return this;
            }
            submissionPayload = submissionPayload.withDestinationId(destinationId);
            return this;
        }

        @Override
        public WithData withAttachments(final List<File> attachments) {
            if (attachments == null) {
                LOGGER.error("Attachments were null");
                return this;
            }
            submissionPayload = submissionPayload.withAttachments(toAttachmentPayloads(attachments));
            return this;
        }

        @Override
        public WithData withAttachment(final File attachment) {
            if (attachment == null) {
                LOGGER.error("Attachment was null");
                return this;
            }
            withAttachments(List.of(attachment));
            return this;
        }

        @Override
        public WithDestination withJsonData(final String jsonData) {
            if (jsonData == null) {
                LOGGER.error("Json data was null");
                return this;
            }
            final ValidationResult validationResult = sender.validateJsonFormat(jsonData);
            if (validationResult.hasError()) {
                LOGGER.error("Data is not in expected json format, please provide valid json {}", validationResult.getError().getMessage());
                return this;
            }
            submissionPayload = submissionPayload.withData(buildDataPayload(jsonData, APPLICATION_JSON));
            return this;
        }


        @Override
        public WithDestination withXmlData(final String xmlData) {
            if (xmlData == null) {
                LOGGER.error("Xml data was null");
                return this;
            }
            final ValidationResult validationResult = sender.validateXmlFormat(xmlData);
            if (validationResult.hasError()) {
                LOGGER.error("Data is not in expected xml format, please provide valid xml {}", validationResult.getError().getMessage());
                return this;
            }
            submissionPayload = submissionPayload.withData(buildDataPayload(xmlData, APPLICATION_XML));
            return this;
        }

        @Override
        public WithEncryptedData withEncryptedAttachments(final Map<UUID, String> encryptedAttachments) {
            if (encryptedAttachments == null) {
                LOGGER.error("Encrypted attachments were null");
                return this;
            }
            submissionPayload = submissionPayload.withAttachments(toAttachmentPayloads(encryptedAttachments));
            return this;
        }

        @Override
        public WithEncryptedData withEncryptedAttachment(final UUID attachmentId, final String encryptedAttachment) {
            if (attachmentId == null) {
                LOGGER.error("AttachmentId was null");
                return this;
            }
            if (encryptedAttachment == null) {
                LOGGER.error("EncryptedAttachment was null");
                return this;
            }
            return withEncryptedAttachments(Map.of(attachmentId, encryptedAttachment));
        }

        @Override
        public WithEncryptedMetadata withEncryptedJsonData(final String encryptedJsonData) {
            if (encryptedJsonData == null) {
                LOGGER.error("Encrypted json data was null");
                return this;
            }
            submissionPayload = submissionPayload.withData(buildEncryptedDataPayload(encryptedJsonData, APPLICATION_JSON));
            return this;
        }

        @Override
        public WithEncryptedMetadata withEncryptedXmlData(final String encryptedXmlData) {
            if (encryptedXmlData == null) {
                LOGGER.error("Encrypted xml data was null");
                return this;
            }
            submissionPayload = submissionPayload.withData(buildEncryptedDataPayload(encryptedXmlData, APPLICATION_XML));
            return this;
        }

        @Override
        public WithDestination withEncryptedMetadata(final String encryptedMetadata) {
            if (encryptedMetadata == null) {
                LOGGER.error("Encrypted metadata was null");
                return this;
            }
            submissionPayload = submissionPayload.withEncryptedMetadata(encryptedMetadata);
            return this;
        }

        @Override
        public Submit withServiceType(final String serviceTypeName, final String leikaKey) {
            if (serviceTypeName == null) {
                LOGGER.error("ServiceType name was null");
                return this;
            }
            if (leikaKey == null) {
                LOGGER.error("LeikaKey was null");
                return this;
            }
            submissionPayload = submissionPayload.withServiceType(buildServiceType(serviceTypeName, leikaKey));
            return this;
        }

        @Override
        public Submit withServiceType(final String serviceTypeName, final String description, final String leikaKey) {
            if (serviceTypeName == null) {
                LOGGER.error("ServiceType name was null");
                return this;
            }
            if (description == null) {
                LOGGER.error("ServiceType description was null");
                return this;
            }
            if (leikaKey == null) {
                LOGGER.error("LeikaKey was null");
                return this;
            }
            submissionPayload = submissionPayload.withServiceType(buildServiceType(serviceTypeName, description, leikaKey));
            return this;
        }

        @Override
        public Optional<SubmitSubmission> submit() {
            return isValidSubmissionData() ? submitStrategy.send(submissionPayload) : Optional.empty();
        }

        private boolean isValidSubmissionData() {
            if (submissionPayload == null) {
                return false;
            } else if (submissionPayload.getDestinationId() == null) {
                LOGGER.error("DestinationId is mandatory, but was null.");
                return false;
            } else if (submissionPayload.getData() == null) {
                LOGGER.error("Data is mandatory, but was null.");
                return false;
            } else if (submissionPayload.getServiceType() == null) {
                LOGGER.error("ServiceType is mandatory, but was null.");
                return false;
            } else {
                return true;
            }
        }

        private Optional<String> retrievePublicKey(final UUID destinationId) {
            final Destination destination = sender.getDestination(destinationId);
            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId, destination.getEncryptionKid());
            final ValidationResult validationResult = sender.validatePublicKey(encryptionKey);
            if (validationResult.hasError() && !environment.isAllowInsecurePublicKey()) {
                LOGGER.error("Invalid public key", validationResult.getError());
                return Optional.empty();
            }
            return Optional.of(encryptionKey.toJSONString());
        }
    }
}
